﻿// Author: Orlys
// Github: https://github.com/Orlys

namespace Yuuna
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using Yuuna.Contracts.Modules;
    using System.Runtime.Loader;
    using System.IO;
    using System.Linq.Expressions;
    using System.Reflection;
     

    public class ModuleManager
    {
        private static volatile ModuleManager s_inst;
        private static object s_lock = new object();
        public static ModuleManager Instance
        {
            get
            {
                if(s_inst is null)
                {
                    lock (s_lock)
                    {
                        if(s_inst is null)
                        {
                            s_inst = new ModuleManager();
                        }
                    }
                }
                return s_inst;

            }
        }

        private const string PATH = "./modules";

        private readonly List< ModuleProxy> _modules;
        private readonly DirectoryInfo _modulesFolder;

        private ModuleManager()
        {
            this._modules = new List<ModuleProxy>();
            this._modulesFolder = new DirectoryInfo(PATH);
            if (!this._modulesFolder.Exists)
            {
                this._modulesFolder.Create();
            }
            else
            {
                this.Scan();
            }
        }

        public IReadOnlyList<ModuleProxy> Modules => this._modules;

        private void Scan()
        {
            foreach (var moduleFolder in this._modulesFolder.EnumerateDirectories())
            {
                if (ModuleProxy.TryCreate(moduleFolder, out var domain))
                    this._modules.Add(domain);
            }
        }

        public void UnloadAll()
        {
            foreach (var m in this._modules)
            {
                m.Unload();
                (m as IDisposable)?.Dispose();
            }
            this._modules.Clear();
        }

        public void ReloadAll()
        {
            this.UnloadAll();
            this.Scan();
        }
    }
}