﻿// Author: Orlys
// Github: https://github.com/Orlys

namespace Yuuna
{
    using System;
    using Yuuna.Contracts.Modules;
    using Yuuna.Contracts.TextSegmention;
    using System.Runtime.Loader;
    using Yuuna.Contracts.Semantics;
    using System.Linq.Expressions;
    using System.IO;
    using System.Linq;
    using System.Runtime.CompilerServices;
    using Yuuna.Contracts.Patterns;

    public sealed class ModuleProxy : IDisposable
    { 
        private volatile ModuleBase _inst;
        private volatile AssemblyLoadContext _loader; 
        private DirectoryInfo _moduleFolder;
        private readonly object _lock = new object();
        public bool Proxied => this._inst is null;

        public string Name => this._inst?.Name;

        public Type Type => this._inst?.GetType();

        private ModuleProxy(DirectoryInfo moduleFolder, AssemblyLoadContext loader, ModuleBase inst)
        {
            this._moduleFolder = moduleFolder;
            this._loader = loader;
            this._inst = inst; 
        }

        public IPatternSet PatternSet => this._inst.Patterns;

        public static bool TryCreate(DirectoryInfo moduleFolder, out ModuleProxy domain)
        {
            if( LoadSingleModuleType(moduleFolder, out var loader, out var inst))
            {
                domain = new ModuleProxy(moduleFolder, loader, inst);
                return true;
            }
            domain = null;
            return false;
        }
         
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        private static bool LoadSingleModuleType(DirectoryInfo moduleFolder, out AssemblyLoadContext loader, out ModuleBase inst)
        {
            loader = new AssemblyLoadContext(moduleFolder.Name, true);

            //var refs = new HashSet<AssemblyName>();
            foreach (var file in moduleFolder.EnumerateFiles("*.dll", SearchOption.AllDirectories))
            {
                try
                {
                    /*var a = */
                    loader.LoadFromAssemblyPath(file.FullName);
                    //foreach (var refA in a.GetReferencedAssemblies()) 
                    //    refs.Add(refA); 
                }
                catch
                {
                }
            }

            var type  = loader.Assemblies
                .SelectMany(x => x.GetTypes())
                .SingleOrDefault(x =>
                    x.IsSubclassOf(typeof(ModuleBase)) &&
                    x.IsPublic &&
                    !x.IsAbstract);

            inst = null;
            if (type is null)
            {
                loader.Unload();
                loader = null;
            }
            else
            {
                try
                {
                    inst = Activator.CreateInstance(type) as ModuleBase;
                }
                catch  
                { 
                } 
            }
            return inst is ModuleBase;
        }

        internal void Initialize(ITextSegmenter textSegmenter, IGroupManager groupManager)
        {
            this._inst.Initialize(textSegmenter, groupManager);
        } 

        public void Reload()
        {
            if(LoadSingleModuleType(this._moduleFolder, out var loader, out var inst))
            {
                lock (this._lock)
                {
                    this.Unload();
                    this._loader = loader;
                    this._inst = inst;
                }
            }
        }

        public void Unload()
        {
            (this._inst as IDisposable)?.Dispose();
            this._inst = null;

            this._loader?.Unload();
            this._loader = null; 
        }

        void IDisposable.Dispose()
        {
            this.Unload();
        }
    }
}